#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "heap.h"

void print_array (double *tab, int size) {
  int i =0;
  for (i =0; i < size - 1; i++) {
    printf ("%f, ", tab[i]);
  }
  printf ("%f\n", tab[i]);
}

int distance_levenshtein (char* str1, char* str2) {
  int len1 = strlen (str1);
  int len2 = strlen (str2);

  int **dist = malloc ( (len1 + 1) * sizeof(int*));
  for (int i = 0; i <= len1; i++) {
    dist[i] = malloc ((len2 + 1) * sizeof (int));
  }

  for (int i = 0; i <= len1; i++) {
    for (int j = 0; j <= len2; j++) {
      dist[i][j] = 0;
    }
  }

  for (int i = 0; i <= len1; i++) {
    dist[i][0] = i;
  }

  for (int i = 0; i <= len2; i++) {
    dist[0][i] = i;
  }

  for (int i = 1; i <= len1; i++) {
    for (int j = 1; j <= len2; j++) {
      int costSubs = 0;
      if (str1[i-1] == str2[j-1]) {
	costSubs = 0;
      }
      else {
	costSubs = 1;
      }

      int v1 = dist[i-1][j]+1;
      int v2 = dist[i][j-1]+1;
      int v3 = dist[i-1][j-1] + costSubs;
      int min = v1 < v2 ? v1 : v2;
      min = min < v3 ? min : v3;
      dist[i][j] = min;
    }
  }

  int res = dist[len1][len2];
  for (int i = 0; i <= len1; i++) {
    free (dist[i]);
  }
  free (dist);
  return res;
}

typedef struct word_distance_ {
  int distance;
  char *mot;
} word_distance_t;

int compareWD (const void *a, const void*b) {
  word_distance_t *wd_a = (word_distance_t *)a;
  word_distance_t *wd_b = (word_distance_t *)b;

 if (wd_a->distance < wd_b->distance) {
   return 1;
 }

 if (wd_a->distance > wd_b->distance) {
   return -1;
 }

 return 0;
}

void* buildWD (const void *a) {
  word_distance_t *wd = (word_distance_t *) a;
  word_distance_t *new_wd = malloc (sizeof(word_distance_t));
  new_wd->distance = wd->distance;
  new_wd->mot = malloc (sizeof(char)*(strlen(wd->mot)+1));
  strcpy(new_wd->mot, wd->mot);
  return (void*)new_wd;
}

int main (int argc, char** argv){

  if (argc != 2) {
    printf ("Usage: %s <chaine>\n", argv[0]);
    return EXIT_SUCCESS;
  }
  /* Mot pour lequel on cherche les plus proches dans le
     dictionnaire */
  char *mot = argv[1];

  int size_dictionnaire = 35;
  char* dictionnaire[] = { "belette", "blaireau", "cerf",
    "couleuvre", "chat", "campagnol", "daim", "ecureuil",
    "founine", "geai", "genette", "grenouille", "herisson",
    "hermine", "lapin", "loup", "lievre", "lynx", "loir",
    "martre", "merle", "mesange", "ours", "pinson", "pivert",
    "processionnaire", "pie", "putois", "renard", "salamandre",
    "sanglier", "sitelle", "taupe", "troglodyte", "vipere" };


  generic_heap_t heap;
  generic_heap_init (&heap, compareWD, buildWD, free);

  for (int i = 0; i < size_dictionnaire; i++) {
    word_distance_t temp;
    temp.distance = distance_levenshtein (mot, dictionnaire[i]);
    temp.mot = malloc(sizeof(char)*(strlen(dictionnaire[i])+1));
    strcpy(temp.mot, dictionnaire[i]);
    generic_heap_insert (&heap, &temp);
  }

  printf ("Mot les plus proche: ");
  while (generic_heap_size (&heap) > 0) {
    word_distance_t* temp = NULL;
    generic_heap_extract (&heap, (void**)&temp);
    if (temp->distance <= 2) {
      printf ("%s, ", temp->mot);
    }
  }
  printf ("\n");

  return EXIT_SUCCESS;
}
